/*
 *  routinesHSMultiResol.cpp
 *
 *
 *  Created by Patrick Heas on 02/12/12.
 *  Copyright 2012 INRIA. All rights reserved.
 *
 */
#ifndef routinesHSMultiResol_h
#define routinesHSMultiResol_h

#include <iostream>
#include <string> 
#include <fstream>
#include <cstdio> 
#include <cstdlib>
#include <cmath>
#include <cassert>

#include "CImg.h"

#include "comprow_double.h"
#include "ilupre_double.h"
#include "icpre_double.h"

#include "diagpre_double.h"
#include "icpre_double.h"
#include "ilupre_double.h"
#include "cgs.h"


using namespace cimg_library;
using namespace std;
 	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/**********************************      Various functions   *************************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
void writeLine(const CImg<double> &save,char * filename);
CImg<double> treshold_minmax(const CImg<double> &in, double treshmin, double treshmax, double flag);
CImg<float> read_aUVFile(char * filename,int dx,int dy,int Dx, int Dy,int ofx, int ofy);
char ** alloc_seqPath(int nbIm);
void read_path(string ifn1, int nbIm, char ** fileIm);
void maj_vecteurs(CImg<double> &X,const CImg<double> & Xin,bool divNul);
CImg<double> divImOrder2(const CImg<double> &im);
CImg<double> expand(const CImg<double> &imOrig,int nbResol);
void expand_dim(int nb_col,int nb_lig,int resol,int *nouv_nb_col,int *nouv_nb_lig);
CImg<short> read_pgm(char *nom,int dx,int dy,int ofx,int ofy);
void write_pgm255(const CImg<double> &in,char * filename, float min, float max);
CImgList<double> alloc_pyramid (int nc, int coarse);
CImgList<double> alloc_pyramid (int nc, int nl, int nslice, int coarse);
void FillChannel(CImg<double> &out , const  CImg<double> & im, int band);
float * CImg2float( const CImg<double> &in);
CImg<double> float2CImg(int dx, int dy, float *in);
void Copy_f(float *im1, float *im2, int nb_col, int nb_line);
CImg<double> fill(const CImg<double> &a,const CImg<double> &b);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/**********************************      Border conditions  *************************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
void setZeroBorder(CImg<double> &a,int ofx, int ofy);
void setZeroBorderOnExpandedRegion(CImg<double> &a,int dxOrig, int dyOrig);
void fixPeriodConditions(int nc,int nl,CImg<double> &A1_i_1j,CImg<double> &A1_i1j,CImg<double> &A1_ij_1,CImg<double> &A1_ij1,
			 CImg<double> &A1_ijnc,CImg<double> &A1_ij0,CImg<double> &A1_inlj,CImg<double> &A1_i0j);
void fixNeumannConditions(int nc,int nl,CImg<double> &A1_i_1j,CImg<double> &A1_ij,CImg<double> &A1_i1j,CImg<double> &A1_ij_1,CImg<double> &A1_ij1);
void excludeWhite(int neta,int lo,CImgList<double> &A_iljl,CImgList<double> &A_i_lj_l,CImgList<double> &A_i_ljl,CImgList<double> &A_ilj_l);	
void fixDirichletConditions(int nc,int nl,CImgList<double> &A_i_lj,CImgList<double> &A_ij_l,CImgList<double> &A_ijl,CImgList<double> &A_ilj,int neta,int lo);	
void fixDirichletConditions(int nc,int nl,CImg<double> &A_i_1j,CImg<double> &A_ij,CImg<double> &A_i1j,CImg<double> &A_ij_1,CImg<double> &A_ij1);	
void fixDirichletConditions_cross(int nc,int nl,CImgList<double> &A_i_lj_l,CImgList<double> &A_i_ljl,CImgList<double> &A_ilj_l,CImgList<double> &A_iljl,int neta,int lo);
void fixDirichletConditions_cross(int nc,int nl,CImg<double> &A_ij,CImgList<double> &A_i_lj_l,CImgList<double> &A_i_ljl,CImgList<double> &A_ilj_l,CImgList<double> &A_iljl,int neta,int lo);
       /*************************************************************************************************************/

	/*************************************************************************************************************/
	/*****************************************Robust  Functions***************************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
void Data_Out_Gen_f_QuasiTV(const CImg<double>&u,const CImg<double>&v,const CImg<double>&fx,const CImg<double>&fy,const CImg<double>&ft,float s,int nb_col,int nb_line,CImg<double> &Data_out);
CImg <double> RobustDataCoef_QuasiTV(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,double sigma);
void Data_Out_Gen_f(const CImg<double>&u,const CImg<double>&v,const CImg<double>&fx,const CImg<double>&fy,const CImg<double>&ft,float sigma,int nb_col,int nb_line,CImg<double> &Data_out);
CImg <double> RobustDataCoef(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,double sigma);
void UpdateRobustWeight_Leclerc(const CImg<double> &UV, CImgList <double> &z, double tau,bool periodic);
void UpdateRobustWeight_QuasiTV(const CImg<double> &UV, CImgList <double> &z, double tau,bool periodic);
void UpdateRobustWeight_QuasiTV(const CImg<double> &UV, CImgList <double> &z_right,  CImgList <double> &z_down,  CImgList <double> &z_rightDown,  CImgList <double> &z_rightUp, CImgList <double> &z_t_right,  CImgList <double> &z_t_down,  CImgList <double> &z_t_rightDown,  CImgList <double> &z_t_rightUp,int neta, int lo, double s, double zeta);

	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/**********************************                  Filters *************************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
CImg<double>  Filtre_2D(const CImg<double>& b, const CImg<double>& filtre);
CImg<double>  Filtre_2D(const CImg<double>& b, const CImg<double>& filtre, bool periodic);
CImg<double>  Filtre_2D_sparse(const CImg<double>& b, const CImg<double>& filtre);
CImg<double>  Filtre_2D_sparse(const CImg<double>& b, const CImg<double>& filtre, bool periodic);
CImg<double> laplacian(const CImg<double> &im, bool periodic);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/*****************************CGS Inversion, log det and trace -1*********************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
int standard_CGS_inversion(int taille, int nombre, double *val,int *row_ptr,int *col_ind, VECTOR_double B, CImg<double> &res,double tol);
int standard_CGS_inversion(int taille, int nombre, double *val,int *row_ptr,int *col_ind, VECTOR_double B, CImg<double> &res);	
double ILU_logDetA(int taille, int nombre, double *val,int *row_ptr,int *col_ind);
double * CGS_randomB_traceInvAA0Fx(int taille, int nombre, double *val,int *row_ptr,int *col_ind,const CImg<double> &IxIx,const CImg<double> &IxIy,const CImg<double> &IyIy,const CImg<double> &data_out,const CImgList<double> &reg_out,const CImg<double> &x, int nbIt, double tol, bool periodic);
double CGS_randomB_traceInvAA0(int taille, int nombre, double *val,int *row_ptr,int *col_ind,const CImg<double> &IxIx,const CImg<double> &IxIy,const CImg<double> &IyIy,const CImg<double> &data_out, int nbIt, double tol);
double CGS_randomB_traceInvA(int taille, int nombre,int nc, int nl, double *val,int *row_ptr,int *col_ind, int nbIt, double tol);	
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/*********************************************Horn and Schunck      ******************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
bool HornAndSchunck(CImg<double> &UV,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,float alpha,const CImg<double> &data_out, bool periodic);
bool HornAndSchunck(CImg<double> &UV,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,float alpha,const CImg<double> &data_out,const CImgList<double> &reg_out, bool periodic);
double DataEnergy(const CImg<double> &U, const CImg<double> &V,const CImg<double> &IxIx,const CImg<double> &IyIy,const CImg<double> &IxIy,const CImg<double> &ItIx,const CImg<double> &ItIy,const CImg<double> &ItIt,const CImg<double> data_out);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/********************************Number element in sparse Hessian Matrix**************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
int NumberNonZeroElementInHessian(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &data_out, const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross, const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp, const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,int scale,int neta, int lo);
int NumberNonZeroElementInHessian_Gen_Rob_BruhnAndAl(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,int scale, int lo);
int NumberNonZeroElementInHessian_Gen_BruhnAndAl(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,int scale, int lo);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/********************************Log determinant of sparse Hessian Matrix*************************************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
double logdetA(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2,const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,int scale, int neta, int lo, int nombre);
double logdetA_Gen_Rob_BruhnAndAl(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,const CImgList <double> &z_right, const CImgList <double> &z_down, const CImgList <double> &z_rightDown, const CImgList <double> &z_rightUp,const CImgList <double> &z_t_right, const CImgList <double> &z_t_down, const CImgList <double> &z_t_rightDown, const CImgList <double> &z_t_rightUp,int scale, int lo, int nombre);
double logdetA_Gen_BruhnAndAl(const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> data_out, const CImg<double> &lambda_S2, const CImg<double> &lambda_S2_cross,int scale, int lo,int nombre);
	/*************************************************************************************************************/

	/*************************************************************************************************************/
	/********************************Evidence of Horn and Schunck beta and alpha parameters    *******************/
	/*************************************************************************************************************/

	/*************************************************************************************************************/
void trueBetaGamma(const CImg<double> &UVinc,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,bool periodic);
double evidence_alphaHS(const CImg<double> &UV,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,const CImgList<double> &reg_out,double alpha,double beta,int nbIt,double tol,double s, bool periodic);
double evidence_betaHS(const CImg<double> &UVinc,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,const CImgList<double> &reg_out,double alpha,int nbIt,double tol,double s,bool periodic);
double * evidence_alphaBetaAndTauHS(const CImg<double> &UVinc,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,double alpha,double beta,int nbIt,double tol,double tau, bool periodic);
double * evidence_alphaBetaAndTauHS_normalized(const CImg<double> &UVinc,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,const CImgList<double> &reg_out,double alpha,double beta,int nbIt,double tol, double tau, bool periodic);
double * evidence_alphaBetaAndTauHS_normalized_QuasiTV(const CImg<double> &UVinc,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,const CImgList<double> &reg_out,double alpha,double beta,int nbIt,double tol, double tau, bool periodic);
double * evidence_alphaBetaAndTauDataAndRegHS_normalized(const CImg<double> &UVinc,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,const CImgList<double> &reg_out,double alpha,double beta,int nbIt,double tol, double tau_d,double tau_r, bool periodic);
double  evidence_tauHS_normalized(const CImg<double> &UVinc,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,double alpha,double beta,int nbIt,double tol, double tau, bool periodic);
double evidence_modelHS(const CImg<double> &UVinc,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,double beta,double gamma, double tau, bool periodic);
double evidence_modelHS_TauVariable(const CImg<double> &UVinc,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,double beta,double gamma, int nbIt,double tol,double tau, bool periodic);
double evidence_modelHS_Mackay(const CImg<double> &UVinc,const CImg<double> &UVtil,const CImg<double> &im_x,const CImg<double> &im_y,const CImg<double> &im_t,const CImg<double> &data_out,const CImgList<double> &reg_out,double beta,double gamma, double tau, bool periodic, int order);

/***********************************************************************************************************************/

#endif
