#include		<stddef.h>
#include		<stdlib.h>
#include		<string.h>
#include		<math.h>
#include		<float.h>
#include		"derivee_spline.h"
#include		<iostream>



////////////////////////////////////PROTOTYPE DES FONCTIONS INTERNES/////////////////////////////////////
// fonction d'allocation
static	double*		allocLinD		(long				n);
static	float*		allocLinF		(long				n);
static	float*		allocVolF		(long				nx,long	ny,long	nz);
static	void		freeLinD		(double				*line);
static	void		freeLinF		(float				*line);
static	void		freeVolF		(float				*volume);
// fonction spline de base (implementation recursive)
// ordre 1 a 7							 
static	double		BsplnWght0		(long				i,
									 double				x);
static	double		BsplnWght1		(long				i,
									 double				x);
static	double		BsplnWght2		(long				i,
									 double				x);
static	double		BsplnWght3		(long				i,
									 double				x);
static	double		BsplnWght4		(long				i,
									 double				x);
static	double		BsplnWght5		(long				i,
									 double				x);
static	double		BsplnWght6		(long				i,
									 double				x);
static	double		BsplnWght7		(long				i,
									 double				x) ;
// derivee des fonctions spline de base									 
static	double		BsplnWght2Deriv		(long				i,
									 double				x) ;

static	double		BsplnWght3Deriv		(long				i,
									 double				x) ;

static	double		BsplnWght4Deriv		(long				i,
									 double				x) ;
									 
static	double		BsplnWght6Deriv                (long				i,
									 double				x);

static	double		BsplnWght3DerivSecond		(long				i,
									 double				x) ;

static	double		BsplnWght4DerivSecond		(long				i,
									 double				x) ;
static	double		BsplnWght6DerivSecond		(long				i,
									 double				x);									 
									 
//fonction d acces memoire
static	long		fold			(long				i,
									 long				n);

static	int			gaussj			(double				a[][3],
									 double				b[][3]);
static	void		getxF2D			(float				*volumeSrc,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*rowDst,
									 long				n);
static	void		getxF2F			(float				*volumeSrc,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 float				*rowDst,
									 long				n);
static	void		getyF2D			(float				*volumeSrc,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*columnDst,
									 long				n);
static	void		getzF2D			(float				*volumeSrc,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*pillarDst,
									 long				n);
static	void		putxD2F			(float				*volumeDst,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*rowSrc,
									 long				n);
static	void		putxF2F			(float				*volumeDst,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 float				*rowSrc,
									 long				n);
static	void		putyD2F			(float				*volumeDst,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*columnSrc,
									 long				n);
static	void		putzD2F			(float				*volumeDst,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*pillarSrc,
									 long				n);
									 
// fonction de convolution
static	void		iirConvolveMirror
									(double				*inPtr,
									 double				*outPtr,
									 long				n,
									 double				gain,
									 double				*realPoles,
									 long				np);
static	int			invertTrsf		(struct trsfStruct	*dirTrsf,
									 struct trsfStruct	*invTrsf);

									 
void		errorReport		(char				*errorMessage)
{
	std::cout<<errorMessage;
}


/*************FONCTION A APPEL***********************/
/************************************************************************/
// obtention des coefficients splines pour la derivation
// inPtr: l'image d'entree
// outPtr:l'image de sortie
// nx,ny,nz: les dimensions de l'image
int			directBsplineMirror
									(float				*inPtr,
									 float				*outPtr,
									 long				nx,
									 long				ny,
									 long				nz,
									 int				degree) {

		double				*data;
		double				realPoles[3];
		double				gain;
		float				*p;
		long				np;
		long				x, y, z;

		p = allocLinF(nx);
		if (p == (float *)NULL) {
		  errorReport("ERROR - Not enough memory in directBsplineMirror");
		  return(ERROR);
		}
		for (z = 0L; (z < nz); z++)
		  for (y = 0L; (y < ny); y++) {
			getxF2F(inPtr, 0L, y, z, nx, ny, p, nx);
			putxF2F(outPtr, 0L, y, z, nx, ny, p, nx);
		  }
		freeLinF(p);
		switch (degree) {
		// recuperation de zi et de co
		  case 0:
		  case 1:
			return(!ERROR);
		  case 2:
			np = 1L;
			realPoles[0] = sqrt(8.0) - 3.0;
			gain = 24.0 - sqrt(512.0);
			break;
		  case 3:
			np = 1L;
			realPoles[0] = sqrt(3.0) - 2.0;
			gain = 12.0 - sqrt(108.0);
			break;
		  case 4:
			np = 2L;
			realPoles[0] = sqrt(664.0 - sqrt(438976.0)) + sqrt(304.0) - 19.0;
			realPoles[1] = sqrt(664.0 + sqrt(438976.0)) - sqrt(304.0) - 19.0;
			gain = (1.0 - realPoles[0]) * (1.0 - realPoles[1]);
			gain *= gain;
			break;
		  case 5:
			np = 2L;
			realPoles[0] = 0.5 * (sqrt(270.0 - sqrt(70980.0)) + sqrt(105.0) - 13.0);
			realPoles[1] = 0.5 * (sqrt(270.0 + sqrt(70980.0)) - sqrt(105.0) - 13.0);
			gain = (1.0 - realPoles[0]) * (1.0 - realPoles[1]);
			gain *= gain;
			break;
		  case 6:
			np = 3L;
			realPoles[0] = -0.488294589303044755130118038883789062112279161239377608394;
			realPoles[1] = -0.081679271076237512597937765737059080653379610398148178525368;
			realPoles[2] = -0.00141415180832581775108724397655859252786416905534669851652709;
			gain = 2.598975999348577818390170516255374207847876853191217652822;
			break;
		  case 7:
			np = 3L;
			realPoles[0] = -0.5352804307964381655424037816816460718339231523426924148812;
			realPoles[1] = -0.122554615192326690515272264359357343605486549427295558490763;
			realPoles[2] = -0.0091486948096082769285930216516478534156925639545994482648003;
			gain = 3.0248282036441843886795463832305782146916878615537002580987;
			break;
		  default:
			errorReport("ERROR - Unknown degree");
			return(ERROR);
		}
		if (nx > 1L) { //application du filtre de coef splines sur toutes les lignes
		  data = allocLinD(nx);
		  if (data == (double *)NULL) {
			errorReport("ERROR - Unable to allocate x data");
			return(ERROR);
		  }
		  for (z = 0L; (z < nz); z++)
			for (y = 0L; (y < ny); y++) {
			  getxF2D(outPtr, 0L, y, z, nx, ny, data, nx);
			  iirConvolveMirror(data, data, nx, gain, realPoles, np);
			  putxD2F(outPtr, 0L, y, z, nx, ny, data, nx);
			}
		  freeLinD(data);
		}
		if (ny > 1L) { //application du filtre de coef splines sur toutes les colonnes
		  data = allocLinD(ny);
		  if (data == (double *)NULL) {
			errorReport("ERROR - Unable to allocate y data");
			return(ERROR);
		  }
		  for (z = 0L; (z < nz); z++)
			for (x = 0L; (x < nx); x++) {
			  getyF2D(outPtr, x, 0L, z, nx, ny, data, ny);
			  iirConvolveMirror(data, data, ny, gain, realPoles, np);
			  putyD2F(outPtr, x, 0L, z, nx, ny, data, ny);
			}
		  freeLinD(data);
		}
		if (nz > 1L) { //application du filtre sur toute la profondeur (cas tridimensionnel)
		  data = allocLinD(nz);
		  if (data == (double *)NULL) {
			errorReport("ERROR - Unable to allocate z data");
			return(ERROR);
		  }
		  for (y = 0L; (y < ny); y++)
			for (x = 0L; (x < nx); x++) {
			  getzF2D(outPtr, x, y, 0L, nx, ny, data, nz);
			  iirConvolveMirror(data, data, nz, gain, realPoles, np);
			  putzD2F(outPtr, x, y, 0L, nx, ny, data, nz);
			}
		  freeLinD(data);
		}
		return(!ERROR);
} /* End of directBsplineMirror */





// derivee suivant x ordre 3,4 et 6
// splinePtr: les coeffeicients spline de l'image
// outPtr: une image de meme taille que celle des coefs pour le gradient
// nx,ny: la taille de l'image

int	gradX	( float				*splinePtr,
		 float				*outPtr,
		long				nx,
		long				ny,
		 int				interpolationDegree,
		float				background) 
{
		long nz=1;
		double				*wtx, *wty, *wtz;
		float				*pi, *pj;
		ptrdiff_t			*fdx, *fdy, *fdz;
		double				weightX[8], weightY[8], weightZ[8];
		ptrdiff_t			foldX[8], foldY[8], foldZ[8];
		struct trsfStruct	trsf, invTrsf;
		double				wx0, wx1, wy0, wy1;
		double				a11, a21, a31;
		double				xO, yO, zO;
		double				xz, yz, zz;
		double				xy, yy, zy;
		double				xIn, yIn, zIn;
		double				xOut, yOut, zOut;
		double				q, qi, qj;
		ptrdiff_t			fx0, fx1, fy0, fy1;
		long				half, width;
		long				nxy = nx * ny;
		long				x, y, z;
		long				i, j, k;
		
		half = (long)interpolationDegree / 2L;
		width = (long)interpolationDegree + 1L;
		switch (interpolationDegree) {
		
		  case 3: // pour l'ordre 3
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) { // pour tous les points
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				 x=(long)xOut;
				 y=(long)yOut;
				 z = (long)zOut;
				 xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
			         weightZ[0] = 1.0;
				 foldZ[0] = (ptrdiff_t)0;
					
				wty = weightY;
				fdy = foldY;
				y -= half;
				for (j = y; (j < (y + width)); j++) { // sur colones avec la spline 
					  *wty++ = BsplnWght3(j, yIn);
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
				}
				wtx = weightX;
				fdx = foldX;

				x -= half;
				for (i = x; (i < (x + width)); i++) { // sur lignes avec la derviee de la spline 
				 *wtx++ = BsplnWght3Deriv(i, xIn);
				 *fdx++ = (ptrdiff_t)fold(i, nx);
				}
				q = 0.0;
				wtz = weightZ;
				fdz = foldZ;
				for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
				 wty = weightY;
				fdy = foldY;
				 pj = splinePtr + *fdz++;
				qj = 0.0;
				 for (j = width; (j-- > 0L);) {
					wtx = weightX;
					fdx = foldX;
					pi = pj + *fdy++;
					qi = 0.0;
					for (i = width; (i-- > 0L);)
					 qi += *wtx++ * (double)*(pi + *fdx++);
					 qj += *wty++ * qi;
				}
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
		  case 4:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				  x = (long)xOut;
				  y = (long)yOut;
				  z = (long)zOut;
				  xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
				     weightZ[0] = 1.0;
				 foldZ[0] = (ptrdiff_t)0;
				 
					wty = weightY;
					fdy = foldY;
					y -= half;

					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght4(j, yIn);
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
					x -= half;
					
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght4Deriv(i, xIn);
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
			case 6:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				  x = (long)xOut;
				  y = (long)yOut;
				  z = (long)zOut;
				 xIn=xOut;
				 yIn=yOut;
				 zIn=zOut;
				
				weightZ[0] = 1.0;
				foldZ[0] = (ptrdiff_t)0;
	
					wty = weightY;
					fdy = foldY;
					y -= half;
					
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght6(j, yIn);
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
					x -= half;
				
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght6Deriv(i, xIn);
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
		  default:
			errorReport("ERROR - Unknown interpolation specification");
			return(ERROR);
		}
		return(!ERROR);
} /* End of affineTransform */



// gradient suivant y; ordre 3,4 et 6 uniquement
int	gradY	( float				*splinePtr,
		 float				*outPtr,
		long				nx,
		long				ny,
		 int				interpolationDegree,
		float				background) 

{
		long nz=1;
		double				*wtx, *wty, *wtz;
		float				*pi, *pj;
		ptrdiff_t			*fdx, *fdy, *fdz;
		double				weightX[8], weightY[8], weightZ[8];
		ptrdiff_t			foldX[8], foldY[8], foldZ[8];
		struct trsfStruct	trsf, invTrsf;
		double				wx0, wx1, wy0, wy1;
		double				a11, a21, a31;
		double				xO, yO, zO;
		double				xz, yz, zz;
		double				xy, yy, zy;
		double				xIn, yIn, zIn;
		double				xOut, yOut, zOut;
		double				q, qi, qj;
		ptrdiff_t			fx0, fx1, fy0, fy1;
		long				half, width;
		long				nxy = nx * ny;
		long				x, y, z;
		long				i, j, k;
		
		half = (long)interpolationDegree / 2L;
		width = (long)interpolationDegree + 1L;
		switch (interpolationDegree) {
		  case 3:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				  x = (long)xOut;
				  y = (long)yOut;
				  z = (long)zOut;
				   xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
					  weightZ[0] = 1.0;
					  foldZ[0] = (ptrdiff_t)0;
					
					wty = weightY;
					fdy = foldY;
					y -= half;
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght3Deriv(j, yIn);
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;

					x -= half;
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght3(i, xIn);
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
		  case 4:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				  x = (long)xOut;
				  y = (long)yOut;
				  z = (long)zOut;
				 xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
				   
					weightZ[0] = 1.0;
					foldZ[0] = (ptrdiff_t)0;
					wty = weightY;
					fdy = foldY;
					y -= half;

					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght4Deriv(j, yIn);
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
					x -= half;

					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght4(i, xIn);
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
			
		case 6:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				
				  x = (long)xOut;
				  y = (long)yOut;
				  z = (long)zOut;
				  xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
	
					  weightZ[0] = 1.0;
					  foldZ[0] = (ptrdiff_t)0;
	
					wty = weightY;
					fdy = foldY;
					y -= half;
					
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght6Deriv(j, yIn);
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
					x -= half;
	
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght6(i, xIn);
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
		  default:
			errorReport("ERROR - Unknown interpolation specification");
			return(ERROR);
		}
		return(!ERROR);
} /* End of affineTransform */

//// derivee double suivant x, ordre 3 et 4 uniquement
int	derivXX	( float				*splinePtr,
		 float				*outPtr,
		long				nx,
		long				ny,
		 int				interpolationDegree,
		float				background) 

{
		long nz=1;
		double				*wtx, *wty, *wtz;
		float				*pi, *pj;
		ptrdiff_t			*fdx, *fdy, *fdz;
		double				weightX[8], weightY[8], weightZ[8];
		ptrdiff_t			foldX[8], foldY[8], foldZ[8];
		struct trsfStruct	trsf, invTrsf;
		double				wx0, wx1, wy0, wy1;
		double				a11, a21, a31;
		double				xO, yO, zO;
		double				xz, yz, zz;
		double				xy, yy, zy;
		double				xIn, yIn, zIn;
		double				xOut, yOut, zOut;
		double				q, qi, qj;
		ptrdiff_t			fx0, fx1, fy0, fy1;
		long				half, width;
		long				nxy = nx * ny;
		long				x, y, z;
		long				i, j, k;
		
		half = (long)interpolationDegree / 2L;
		width = (long)interpolationDegree + 1L;
		switch (interpolationDegree) {
		  case 3:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				 
				  x = (long)xOut;
				  y = (long)yOut;
				  z = (long)zOut;
				 xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
			
					  weightZ[0] = 1.0;
					  foldZ[0] = (ptrdiff_t)0;
					
					wty = weightY;
					fdy = foldY;
	
					if ((yIn < 0.0) && ((double)y != yIn))
					  y--;
					y -= half;
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght3(j, yIn);
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
				
					x -= half;
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght3DerivSecond(i, xIn);
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
		  case 4:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
		
				  x = (long)xOut;
				  y = (long)yOut;
				  z = (long)zOut;
				  xIn=xOut;
				  yIn=yOut;
				  zIn=zOut;
		
					weightZ[0] = 1.0;
					foldZ[0] = (ptrdiff_t)0;
					wty = weightY;
					fdy = foldY;
					y -= half;
					
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght4(j, yIn); 
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
					x -= half;
					
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght4DerivSecond(i, xIn);
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
			case 6:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				 x=(long)xOut;
				 y=(long)yOut;
				 z = (long)zOut;
				 xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
			
					  weightZ[0] = 1.0;
					  foldZ[0] = (ptrdiff_t)0;
	
					wty = weightY;
					fdy = foldY;
					y -= half;
					
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght6(j, yIn); 
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
					x -= half;
					
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght6DerivSecond(i, xIn); 
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			}break;
			
		  default:
			errorReport("ERROR - Unknown interpolation specification");
			return(ERROR);
		}
		return(!ERROR);
} /* End of affineTransform */



// derivee double suivant y, ordre 3 et 4 uniquement
int	derivYY	( float				*splinePtr,
		 float				*outPtr,
		long				nx,
		long				ny,
		 int				interpolationDegree,
		float				background) 

{
		long nz=1;
		double				*wtx, *wty, *wtz;
		float				*pi, *pj;
		ptrdiff_t			*fdx, *fdy, *fdz;
		double				weightX[8], weightY[8], weightZ[8];
		ptrdiff_t			foldX[8], foldY[8], foldZ[8];
		struct trsfStruct	trsf, invTrsf;
		double				wx0, wx1, wy0, wy1;
		double				a11, a21, a31;
		double				xO, yO, zO;
		double				xz, yz, zz;
		double				xy, yy, zy;
		double				xIn, yIn, zIn;
		double				xOut, yOut, zOut;
		double				q, qi, qj;
		ptrdiff_t			fx0, fx1, fy0, fy1;
		long				half, width;
		long				nxy = nx * ny;
		long				x, y, z;
		long				i, j, k;
		
		half = (long)interpolationDegree / 2L;
		width = (long)interpolationDegree + 1L;
		switch (interpolationDegree) {
		  case 3:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				 
				 x=(long)xOut;
				 y=(long)yOut;
				 z = (long)zOut;
				 xIn=xOut;
				 yIn=yOut;
				 zIn=zOut;

				 weightZ[0] = 1.0;
				foldZ[0] = (ptrdiff_t)0;
					
					wty = weightY;
					fdy = foldY;
					
					y -= half;
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght3DerivSecond(j, yIn);
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
			
					x -= half;
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght3(i, xIn);
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
		  case 4:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				x=(long)xOut;
				 y=(long)yOut;
				 z = (long)zOut;
				 xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
					weightZ[0] = 1.0;
					foldZ[0] = (ptrdiff_t)0;
					wty = weightY;
					fdy = foldY;
					y -= half;
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght4DerivSecond(j, yIn);
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
					x -= half;
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght4(i, xIn);
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
			
		case 6:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				 x=(long)xOut;
				 y=(long)yOut;
				 z = (long)zOut;
				 xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
					  weightZ[0] = 1.0;
					  foldZ[0] = (ptrdiff_t)0;
	
					wty = weightY;
					fdy = foldY;
					y -= half;
					
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght6DerivSecond(j, yIn); 
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
					x -= half;
					
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght6(i, xIn); 
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			}break;
			
		  default:
			errorReport("ERROR - Unknown interpolation specification");
			return(ERROR);
		}
		return(!ERROR);
} /* End of affineTransform */

// derivation croise d'ordre 3 et 4 uniquement
int	derivXY	( float				*splinePtr,
		 float				*outPtr,
		long				nx,
		long				ny,
		 int				interpolationDegree,
		float				background) 

{
		long nz=1;
		double				*wtx, *wty, *wtz;
		float				*pi, *pj;
		ptrdiff_t			*fdx, *fdy, *fdz;
		double				weightX[8], weightY[8], weightZ[8];
		ptrdiff_t			foldX[8], foldY[8], foldZ[8];
		struct trsfStruct	trsf, invTrsf;
		double				wx0, wx1, wy0, wy1;
		double				a11, a21, a31;
		double				xO, yO, zO;
		double				xz, yz, zz;
		double				xy, yy, zy;
		double				xIn, yIn, zIn;
		double				xOut, yOut, zOut;
		double				q, qi, qj;
		ptrdiff_t			fx0, fx1, fy0, fy1;
		long				half, width;
		long				nxy = nx * ny;
		long				x, y, z;
		long				i, j, k;
		
		half = (long)interpolationDegree / 2L;
		width = (long)interpolationDegree + 1L;
		switch (interpolationDegree) {
		  case 3:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
					 
				 x=(long)xOut;
				 y=(long)yOut;
				 z = (long)zOut;
				 xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
				   
					  weightZ[0] = 1.0;
					  foldZ[0] = (ptrdiff_t)0;
					
					wty = weightY;
					fdy = foldY;
				
					y -= half;
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght3Deriv(j, yIn);
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					
					x -= half;
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght3Deriv(i, xIn);
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
		  case 4:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				
				 x=(long)xOut;
				 y=(long)yOut;
				 z = (long)zOut;
				 xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
				   
				weightZ[0] = 1.0;
				foldZ[0] = (ptrdiff_t)0;
					wty = weightY;
					fdy = foldY;
					y -= half;
	
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght4Deriv(j, yIn);
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
					x -= half;
	
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght4Deriv(i, xIn);
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			  }
			break;
		
		case 6:
			zO=0;yO=0;xO=0;zOut=0;
			  for (yOut = -yO; (yOut < ((double)ny - yO)); yOut += 1.0) {
				for (xOut = -xO; (xOut < ((double)nx - xO)); xOut += 1.0) {
				x=(long)xOut;
				 y=(long)yOut;
				 z = (long)zOut;
				 xIn=xOut;
				  yIn=yOut;
				   zIn=zOut;
				 
					  weightZ[0] = 1.0;
					  foldZ[0] = (ptrdiff_t)0;
	
					wty = weightY;
					fdy = foldY;
					y -= half;
					
					for (j = y; (j < (y + width)); j++) {
					  *wty++ = BsplnWght6Deriv(j, yIn); 
					  *fdy++ = (ptrdiff_t)(fold(j, ny) * nx);
					}
					wtx = weightX;
					fdx = foldX;
					x -= half;
					
					for (i = x; (i < (x + width)); i++) {
					  *wtx++ = BsplnWght6Deriv(i, xIn); 
					  *fdx++ = (ptrdiff_t)fold(i, nx);
					}
					q = 0.0;
					wtz = weightZ;
					fdz = foldZ;
					for (k = (nz == 1L) ? (1L) : (width); (k-- > 0L);) {
					  wty = weightY;
					  fdy = foldY;
					  pj = splinePtr + *fdz++;
					  qj = 0.0;
					  for (j = width; (j-- > 0L);) {
						wtx = weightX;
						fdx = foldX;
						pi = pj + *fdy++;
						qi = 0.0;
						for (i = width; (i-- > 0L);)
						  qi += *wtx++ * (double)*(pi + *fdx++);
						qj += *wty++ * qi;
					  }
					  q += *wtz++ * qj;
					}
				  *outPtr++ = (float)q;
				}
			}break;

		  default:
			errorReport("ERROR - Unknown interpolation specification");
			return(ERROR);
		}
		return(!ERROR);
} /* End of affineTransform */





//

/************************************************************************/
/************************************************************************/
static	double*		allocLinD		(long				n) {

		return((double *)malloc((size_t)n * sizeof(double)));
} /* End of allocLinD */

/************************************************************************/
static	float*		allocLinF		(long				n) {

		return((float *)malloc((size_t)n * sizeof(float)));
} /* End of allocLinF */

/************************************************************************/
static	float*		allocVolF		(long				nx,
									 long				ny,
									 long				nz) {

		return((float *)malloc((size_t)(nx * ny * nz) * sizeof(float)));
} /* End of allocVolF */

/************************************************************************/
static	double		BsplnWght0		(long				i,
									 double				x) {

		x -= (double)i;
		return((x >= 0.5) ? (0.0) : ((x < -0.5) ? (0.0) : (1.0)));
} /* End of BsplnWght0 */

/************************************************************************/
static	double		BsplnWght1		(long				i,
									 double				x) {

		x = fabs(x - (double)i);
		return((x > 1.0) ? (0.0) : (1.0 - x));
} /* End of BsplnWght1 */

/************************************************************************/
static	double		BsplnWght2		(long				i,
									 double				x) {

		x = fabs(x - (double)i);
		if (x < 0.5)
		  return(0.75 - x * x);
		if (x < 1.5) {
		  x = 1.5 - x;
		  return(0.5 * x * x);
		}
		return(0.0);
} /* End of BsplnWght2 */

/************************************************************************/
static	double		BsplnWght3		(long				i,
									 double				x) {

		x = fabs(x - (double)i);
		if (x < 1.0)
		  return((x * x * (x - 2.0) * 3.0 + 4.0) * (1.0 / 6.0));
		if (x < 2.0) {
		  x = 2.0 - x;
		  return(x * x * x * (1.0 / 6.0));
		}
		return(0.0);
} /* End of BsplnWght3 */

/************************************************************************/
static	double		BsplnWght4		(long				i,
									 double				x) {

		x = fabs(x - (double)i);
		if (x < 0.5) {
		  x *= x;
		  return(x * (x * 0.25 - 0.625) + 115.0 / 192.0);
		}
		if (x < 1.5)
		  return(x * (x * (x * (5.0 / 6.0 - x * (1.0 / 6.0)) - 1.25) + 5.0 / 24.0) + 55.0
			/ 96.0);
		if (x < 2.5) {
		  x -= 2.5;
		  x *= x;
		  return(x * x * (1.0 / 24.0));
		}
		return(0.0);
} /* End of BsplnWght4 */

/************************************************************************/
static	double		BsplnWght5		(long				i,
									 double				x) {

		double				f;

		x = fabs(x - (double)i);
		if (x < 1.0) {
		  f = x * x;
		  return(f * (f * (0.25 - x * (1.0 / 12.0)) - 0.5) + 0.55);
		}
		if (x < 2.0)
		  return(x * (x * (x * (x * (x * (1.0 / 24.0) - 0.375) + 1.25) - 1.75) + 0.625)
			+ 0.425);
		if (x < 3.0) {
		  f = 3.0 - x;
		  x = f * f;
		  return(f * x * x * (1.0 / 120.0));
		}
		return(0.0);
} /* End of BsplnWght5 */

/************************************************************************/
static	double		BsplnWght6		(long				i,
									 double				x) {

		x = fabs(x - (double)i);
		if (x < 0.5) {
		  x *= x;
		  return(x * (x * (7.0 / 48.0 - x * (1.0 / 36.0)) - 77.0 / 192.0) + 5887.0 / 11520.0);
		}
		if (x < 1.5)
		  return(x * (x * (x * (x * (x * (x * (1.0 / 48.0) - 7.0 / 48.0) + 0.328125)
			- 35.0 / 288.0) - 91.0 / 256.0) - 7.0 / 768.0) + 7861.0 / 15360.0);
		if (x < 2.5)
		  return(x * (x * (x * (x * (x * (7.0 / 60.0 - x * (1.0 / 120.0)) - 0.65625)
			+ 133.0 / 72.0) - 2.5703125) + 1267.0 / 960.0) + 1379.0 / 7680.0);
		if (x < 3.5) {
		  x -= 3.5;
		  x *= x * x;
		  return(x * x * (1.0 / 720.0));
		}
		return(0.0);
} /* End of BsplnWght6 */

/************************************************************************/
static	double		BsplnWght7		(long				i,
									 double				x) {

		double				f;

		x = fabs(x - (double)i);
		if (x < 1.0) {
		  f = x * x;
		  return(f * (f * (f * (x * (1.0 / 144.0) - 1.0 / 36.0) + 1.0 / 9.0) - 1.0 / 3.0)
			+ 151.0 / 315.0);
		}
		if (x < 2.0)
		  return(x * (x * (x * (x * (x * (x * (0.05 - x * (1.0 / 240.0)) - 7.0 / 30.0) + 0.5)
			- 7.0 / 18.0) - 0.1) - 7.0 / 90.0) + 103.0 / 210.0);
		if (x < 3.0)
		  return(x * (x * (x * (x * (x * (x * (x * (1.0 / 720.0) - 1.0 / 36.0) + 7.0 / 30.0)
			- 19.0 / 18.0) + 49.0 / 18.0) - 23.0 / 6.0) + 217.0 / 90.0) - 139.0 / 630.0);
		if (x < 4.0) {
		  f = 4.0 - x;
		  x = f * f * f;
		  return(x * x * f * (1.0 / 5040.0));
		}
		return(0.0);
} /* End of BsplnWght7 */

static	double		BsplnWght2Deriv		(long				i,
									 double				x) {

		return ( BsplnWght1(i,x+0.5) -  BsplnWght1(i,x-0.5) );
} /* End of BsplnWght3 */

static	double		BsplnWght3Deriv		(long				i,
									 double				x) {

		return ( BsplnWght2(i,x+0.5) -  BsplnWght2(i,x-0.5) );
} /* End of BsplnWght3 */

static	double		BsplnWght4Deriv		(long				i,
									 double				x) {

		return ( BsplnWght3(i,x+0.5) -  BsplnWght3(i,x-0.5) );
} /* End of BsplnWght3 */

static	double		BsplnWght5Deriv		(long				i,
									 double				x) {

		return ( BsplnWght4(i,x+0.5) -  BsplnWght4(i,x-0.5) );
} /* End of BsplnWght3 */


static	double	BsplnWght6Deriv                (long				i,
									 double				x) {
		return ( BsplnWght5(i,x+0.5) -  BsplnWght5(i,x-0.5) );
} /* End of BsplnWght3 */									 

static	double		BsplnWght3DerivSecond		(long				i,
									 double				x) {

		return ( BsplnWght2Deriv(i,x+0.5) -  BsplnWght2Deriv(i,x-0.5) );
} /* End of BsplnWght3 */

static	double		BsplnWght4DerivSecond		(long				i,
									 double				x) {

		return ( BsplnWght3Deriv(i,x+0.5) -  BsplnWght3Deriv(i,x-0.5) );
} /* End of BsplnWght3 */
static	double		BsplnWght6DerivSecond		(long				i,
									 double				x) {

		return ( BsplnWght5Deriv(i,x+0.5) -  BsplnWght5Deriv(i,x-0.5) );
} /* End of BsplnWght3 */




/************************************************************************/
static	long		fold			(long				i,
									 long				n) {

		ldiv_t				modOp;
		long				n2;

		i = labs(i);
		if (i < n)
		  return(i);
		if (n == 1L)
		  return(0L);
		n2 = (n << 1L) - 2L;
		modOp = ldiv(i, n2);
		return((modOp.rem < n) ? (modOp.rem) : (n2 - modOp.rem));
} /* End of fold */

/************************************************************************/
static	void		freeLinD		(double				*line) {

		free(line);
} /* End of freeLinD */

/************************************************************************/
static	void		freeLinF		(float				*line) {

		free(line);
} /* End of freeLinF */

/************************************************************************/
static	void		freeVolF		(float				*volume) {

		free(volume);
} /* End of freeVolF */

/************************************************************************/

/************************************************************************/
static	void		getxF2D			(float				*volumeSrc,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*rowDst,
									 long				n) {

		double				*end;

		volumeSrc += (ptrdiff_t)(x + nx * (y + ny * z));
		end = rowDst + (ptrdiff_t)n;
		while (rowDst < end)
		  *rowDst++ = (double)*volumeSrc++;
} /* End of getxF2D */

/************************************************************************/
static	void		getxF2F			(float				*volumeSrc,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 float				*rowDst,
									 long				n) {

		volumeSrc += (ptrdiff_t)(x + nx * (y + ny * z));
		rowDst = (float *)memcpy(rowDst, volumeSrc, (size_t)n * sizeof(float));
} /* End of getxF2F */

/************************************************************************/
static	void		getyF2D			(float				*volumeSrc,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*columnDst,
									 long				n) {

		double				*end;

		volumeSrc += (ptrdiff_t)(x + nx * (y + ny * z));
		for (end = columnDst + (ptrdiff_t)n; (columnDst < end); volumeSrc += (ptrdiff_t)nx)
		  *columnDst++ = (double)*volumeSrc;
} /* End of getyF2D */

/************************************************************************/
static	void		getzF2D			(float				*volumeSrc,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*pillarDst,
									 long				n) {

		double				*end;
		long				nxny = nx * ny;

		volumeSrc += (ptrdiff_t)(x + nx * y + nxny * z);
		for (end = pillarDst + (ptrdiff_t)n; (pillarDst < end); volumeSrc += (ptrdiff_t)nxny)
		  *pillarDst++ = (double)*volumeSrc;
} /* End of getzF2D */

/************************************************************************/
static	void		iirConvolveMirror
									(double				*inPtr,
									 double				*outPtr,
									 long				n,
									 double				gain,
									 double				*realPoles,
									 long				np) {

		double				*p, *q, *r;
		double				tolerance = log10((double)FLT_EPSILON);
		double				x0, pole;
		long				n2;
		long				i, j, k;

		p = inPtr;
		q = outPtr;
		r = outPtr + (ptrdiff_t)n;
		while (q < r)
		  *q++ = *p++ * gain;
		if (n == 1L)
		  return;
		n2 = 2L * (n - 1L);
		i = np;
		while (i-- > 0L) {
		  pole = *realPoles++;
		  j = (long)ceil(tolerance / log10(fabs(pole)));
		  k = j - n2 * (j / n2);
		  j -= k;
		  if (k < n) {
			p = outPtr + (ptrdiff_t)k;
			x0 = *p;
		  }
		  else {
			k = n2 - k;
			p = outPtr + (ptrdiff_t)k;
			x0 = *p;
			while (++p < r)
			  x0 = pole * x0 + *p;
			p--;
		  }
		  while (--p >= outPtr)
			x0 = pole * x0 + *p;
		  while (j > 0L) {
			p++;
			while (++p < r)
			  x0 = pole * x0 + *p;
			p--;
			while (--p >= outPtr)
			  x0 = pole * x0 + *p;
			j -= n2;
		  }
		  q = p++;
		  *p++ = x0;
		  x0 = *(q++ + (ptrdiff_t)n);
		  while (p < r)
			*p++ += *q++ * pole;
		  *q = (2.0 * *q - x0) / (1.0 - pole * pole);
		  p--;
		  while (--q >= outPtr)
			*q += *p-- * pole;
		}
} /* End of iirConvolveMirror */

/************************************************************************/
static	void		putxD2F			(float				*volumeDst,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*rowSrc,
									 long				n) {

		double				*end;

		volumeDst += (ptrdiff_t)(x + nx * (y + ny * z));
		end = rowSrc + (ptrdiff_t)n;
		while (rowSrc < end)
		  *volumeDst++ = (float)*rowSrc++;
} /* End of putxD2F */

/************************************************************************/
static	void		putxF2F			(float				*volumeDst,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 float				*rowSrc,
									 long				n) {

		volumeDst += (ptrdiff_t)(x + nx * (y + ny * z));
		volumeDst = (float *)memcpy(volumeDst, rowSrc, (size_t)n * sizeof(float));
} /* End of putxF2F */

/************************************************************************/
static	void		putyD2F			(float				*volumeDst,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*columnSrc,
									 long				n) {

		double				*end;

		volumeDst += (ptrdiff_t)(x + nx * (y + ny * z));
		for (end = columnSrc + (ptrdiff_t)n; (columnSrc < end); volumeDst += (ptrdiff_t)nx)
		  *volumeDst = (float)*columnSrc++;
} /* End of putyD2F */

/************************************************************************/
static	void		putzD2F			(float				*volumeDst,
									 long				x,
									 long				y,
									 long				z,
									 long				nx,
									 long				ny,
									 double				*pillarSrc,
									 long				n) {

		double				*end;
		long				nxny = nx * ny;

		volumeDst += (ptrdiff_t)(x + nx * y + nxny * z);
		for (end = pillarSrc + (ptrdiff_t)n; (pillarSrc < end); volumeDst += (ptrdiff_t)nxny)
		  *volumeDst = (float)*pillarSrc++;
} /* End of putzD2F */

